from ...math import log, sinh, sqrt


def langevin(stretch, mu, N):
    """Langevin model given by the free energy of a single chain as a function
    of the stretch (assuming a complex valued logarithm). The inverse
    Langevin function is defined by a Padé approximation."""

    x = stretch / sqrt(N)
    L = x * (3 - x ** 2) / (1 - x ** 2)

    return mu * N * (x * L + log(L / sinh(L)))


def langevin2(stretch, mu, N):
    """Langevin model given by the free energy of a single chain as a function
    of the stretch (assuming a complex valued logarithm). The inverse Langevin
    function is defined by a Padé approximation.

    **Note**: This function is optimized for fast gradient evaluation but
    cannot be used to calculate the actual free energy - only its derivative
    w.r.t. the stretch gives a real-valued result for the region of interest
    `N > stretch ** 2`."""

    return mu * (stretch ** 2 / 2 - N * log(stretch ** 2 - N))


def gauss(stretch, mu):
    """Gaussian model given by the free energy
    of a single chain as a function of the stretch."""

    return 3 * mu / 2 * (stretch ** 2 - 1)


def linear(stretch, mu):
    """Linear model given by the free energy
    of a single chain as a function of the stretch."""

    return mu * (stretch - 1)



from multiprocessing import cpu_count

import numpy as np

from . import Material, Variable
from .math import det, horzcat, vertcat, zeros


class ThreeFieldVariation:
    def __init__(self, material):
        self.material = material
        p = Variable("p", 1, 1)
        J = Variable("J", 1, 1)
        self.x = [self.material.x[0], p, J]
        self.W = Material(self.x, self._fun)
        self.function = self.W.function
        self.gradient = self.W.gradient
        self.hessian = self.W.hessian

    def _fun(self, x):
        F, p, J = x
        detF = det(F)
        Fmod = (J / detF) ** (1 / 3) * F
        return self.material.fun(Fmod, **self.material.kwargs) + p * (detF - J)


class MaterialHyperelastic:
    def __init__(self, fun, **kwargs):
        F = Variable("F", 3, 3)
        self.x = [F]
        self.fun = fun
        self.kwargs = kwargs
        self.W = Material(self.x, self._fun_wrapper, kwargs=self.kwargs)
        self.function = self.W.function
        self.gradient = self.W.gradient
        self.hessian = self.W.hessian

    def _fun_wrapper(self, x, **kwargs):
        return self.fun(x[0], **kwargs)


class MaterialHyperelasticPlaneStrain:
    def __init__(self, fun, **kwargs):
        F = Variable("F", 2, 2)
        self.x = [F]
        self.fun = fun
        self.kwargs = kwargs
        self.W = Material(self.x, self._fun_wrapper, kwargs=self.kwargs)
        self.function = self.W.function
        self.gradient = self.W.gradient
        self.hessian = self.W.hessian

    def _fun_wrapper(self, x, **kwargs):
        F = horzcat(vertcat(x[0], zeros(1, 2)), zeros(3, 1))
        F[2, 2] = 1  # fixed thickness ratio `h / H = 1`
        return self.fun(F, **kwargs)


class MaterialHyperelasticPlaneStressIncompressible(MaterialHyperelasticPlaneStrain):
    def __init__(self, fun, **kwargs):
        super().__init__(fun, **kwargs)

    def _fun_wrapper(self, x, **kwargs):
        F = horzcat(vertcat(x[0], zeros(1, 2)), zeros(3, 1))
        F[2, 2] = 1 / det(x[0])  # thickness ratio `h / H = 1 / (a / A)`
        return self.fun(F, **kwargs)


class MaterialComposite:
    def __init__(self, materials):
        "Composite Material as a sum of a list of materials."
        self.materials = materials
        self.fun = self.composite

    def composite(self):
        "Dummy function for plot title."
        return

    def function(self, x, **kwargs):
        fun = [m.function(x, **kwargs) for m in self.materials]
        return [np.sum([f[a] for f in fun], 0) for a in range(len(fun[0]))]

    def gradient(self, x, **kwargs):
        grad = [m.gradient(x, **kwargs) for m in self.materials]
        return [np.sum([g[a] for g in grad], 0) for a in range(len(grad[0]))]

    def hessian(self, x, **kwargs):
        hess = [m.hessian(x, **kwargs) for m in self.materials]
        return [np.sum([h[a] for h in hess], 0) for a in range(len(hess[0]))]



from multiprocessing import cpu_count

import numpy as np
import casadi as ca

from ._apply import apply


class Material:
    def __init__(self, x, fun, args=(), kwargs={}, compress=False):

        self.x = x
        self._fun = fun

        self.args = args
        self.kwargs = kwargs

        self._f = self._fun(self.x, *self.args, **self.kwargs)

        _h_diag = []

        self._h = []
        self._g = []

        self.jacobian = self.gradient

        # generate list of diagonal hessian entries and gradients
        for y in self.x:
            _hy, _gy = ca.hessian(self._f, y)
            _h_diag.append(_hy)
            self._g.append(_gy)

        # generate upper-triangle of hessian
        for i, g in enumerate(self._g):
            for j, y in enumerate(self.x):
                if j >= i:
                    if i != j:
                        self._h.append(ca.jacobian(g, y))
                    else:
                        self._h.append(_h_diag[i])

        # generate casADi function objects
        self._function = ca.Function("f", self.x, [self._f])
        self._gradient = ca.Function("g", self.x, self._g)
        self._hessian = ca.Function("h", self.x, self._h)

        # generate indices
        self._idx_gradient = [y.shape for y in x]
        self._idx_hessian = []

        if compress:
            for i in range(len(self._idx_gradient)):
                if np.all(np.array(self._idx_gradient[i]) == 1):
                    self._idx_gradient[i] = ()

        for i in range(len(self._idx_gradient)):
            a = self._idx_gradient[i]

            for j in range(len(self._idx_gradient)):
                b = self._idx_gradient[j]

                if j >= i:
                    self._idx_hessian.append((*a, *b))

    def function(self, x, threads=cpu_count()):
        "Return the function."
        return apply(
            x,
            fun=self._function,
            x_shape=self._idx_gradient,
            fun_shape=[()],
            threads=threads,
        )

    def gradient(self, x, threads=cpu_count()):
        "Return list of gradients."
        return apply(
            x,
            fun=self._gradient,
            x_shape=self._idx_gradient,
            fun_shape=self._idx_gradient,
            threads=threads,
        )

    def hessian(self, x, threads=cpu_count()):
        "Return upper-triangle entries of hessian."
        return apply(
            x,
            fun=self._hessian,
            x_shape=self._idx_gradient,
            fun_shape=self._idx_hessian,
            threads=threads,
        )


class MaterialTensor:
    def __init__(self, x, fun, args=(), kwargs={}, compress=False):

        self.x = x
        self._fun = fun

        self.args = args
        self.kwargs = kwargs

        self._f = self._fun(self.x, *self.args, **self.kwargs)
        self._g = [ca.jacobian(self._f, y) for y in self.x]

        # alias
        self.jacobian = self.gradient

        # generate casADi function objects
        self._function = ca.Function("f", self.x, [self._f])
        self._gradient = ca.Function("g", self.x, self._g)

        # generate indices
        self._idx_function = [y.shape for y in x]
        self._idx_gradient = []

        if compress:
            for i in range(len(self._idx_function)):
                if np.all(np.array(self._idx_function[i]) == 1):
                    self._idx_function[i] = ()

        for i in range(len(self._idx_function)):
            a = self._idx_function[i]

            for j in range(len(self._idx_function)):
                b = self._idx_function[j]

                self._idx_gradient.append((*a, *b))

    def function(self, x, threads=cpu_count()):
        "Return the function."
        return apply(
            x,
            fun=self._function,
            x_shape=self._idx_function,
            fun_shape=self._idx_function,
            threads=threads,
        )

    def gradient(self, x, threads=cpu_count()):
        "Return list of gradients."
        return apply(
            x,
            fun=self._gradient,
            x_shape=self._idx_function,
            fun_shape=self._idx_gradient,
            threads=threads,
        )



from casadi import (
    # tensor and vector operations
    det,
    inv,
    dot,
    transpose,
    trace,
    diag,
    adj,
    cofactor,
    cross,
    times,
    eig_symbolic,
    ldl,
    # trig
    sin,
    cos,
    tan,
    sinh,
    cosh,
    tanh,
    asin,
    acos,
    atan,
    atan2,
    asinh,
    acosh,
    atanh,
    # math
    exp,
    log,
    sqrt,
    sum1,
    sum2,
    sumsqr,
    pi,
    fabs,
    linspace,
    erf,
    erfinv,
    norm_1,
    sign,
    fmin,
    fmax,
    mmin,
    mmax,
    find,
    #
    if_else,
    logic_and,
    logic_or,
    logic_not,
    floor,
    ceil,
    SX,
    DM,
    MX,
    #
    vertcat,
    horzcat,
    vertsplit,
    horzsplit,
    reshape,
)

eye = SX.eye
ones = SX.ones
zeros = SX.zeros


def invariants(T):

    I1 = trace(T)
    I2 = (I1 ** 2 - trace(T @ T)) / 2
    I3 = det(T)

    return I1, I2, I3


def eigvals(T, eps=1e-5):

    T[0, 0] += eps
    T[1, 1] -= eps

    wT = eig_symbolic(T)

    return wT


def cof(T):

    return det(T) * transpose(inv(T))


def sym(T):

    return (T + transpose(T)) / 2



from .__about__ import __version__

import casadi

Variable = casadi.SX.sym

from ._material import (
    Material,
    Material as MaterialScalar,
    MaterialTensor,
)
from . import models
from . import math
from ._templates import (
    ThreeFieldVariation,
    MaterialHyperelastic,
    MaterialComposite,
    MaterialHyperelasticPlaneStrain,
    MaterialHyperelasticPlaneStressIncompressible,
)
from ._lab import Lab

__all__ = [
    "__version__",
]



from functools import wraps
from copy import deepcopy

from ..math import det


def isochoric_volumetric_split(fun):
    """Apply the material formulation only on the isochoric part of the
    multiplicative split of the deformation gradient. Optionally, if
    the bulk keyword is passed, add a volumetric term."""

    @wraps(fun)
    def apply_iso(*args, **kwargs):
        F = args[0]
        J = det(F)
        F_iso = J ** (-1 / 3) * F

        fun_args = args[1:]
        fun_kwargs = deepcopy(kwargs)

        if "bulk" in kwargs.keys():
            _ = fun_kwargs.pop("bulk")

        W = fun(F_iso, *fun_args, **fun_kwargs)

        if "bulk" in kwargs.keys():
            W += volumetric(J, kwargs["bulk"])

        return W

    return apply_iso


def volumetric(J, bulk):
    return bulk * (J - 1) ** 2 / 2



import numpy as np

from matadi import Variable, Material, MaterialTensor
from matadi.math import det, transpose, trace, invariants, sqrt


def neohooke(x, mu=1.0, bulk=200.0):
    """Strain energy density function of nearly-incompressible
    Neo-Hookean isotropic hyperelastic material formulation."""

    F = x[0]
    C = transpose(F) @ F

    I1, I2, I3 = invariants(C)
    J = sqrt(I3)

    I1_iso = I3 ** (-1 / 3) * trace(C)

    return mu * (I1_iso - 3) + bulk * (J - 1) ** 2 / 2


def test_simple():

    # variables
    F = Variable("F", 3, 3)

    # data
    FF = np.random.rand(3, 3, 8, 1000)
    for a in range(3):
        FF[a, a] += 1

    # init Material
    W = Material(
        x=[F],
        fun=neohooke,
        kwargs={"mu": 1.0, "bulk": 10.0},
    )

    W0 = W.function([FF])
    dW = W.gradient([FF])
    DW = W.hessian([FF])

    # dW and DW are always lists...
    assert W0[0].shape == (8, 1000)
    assert dW[0].shape == (3, 3, 8, 1000)
    assert DW[0].shape == (3, 3, 3, 3, 8, 1000)

    # check output of parallel evaluation
    W0 = W.function([FF], threads=2)
    dW = W.gradient([FF], threads=2)
    DW = W.hessian([FF], threads=2)

    assert W0[0].shape == (8, 1000)
    assert dW[0].shape == (3, 3, 8, 1000)
    assert DW[0].shape == (3, 3, 3, 3, 8, 1000)


def test_tensor():

    # variables
    F = Variable("F", 3, 3)
    p = Variable("p", 1, 1)

    # data
    FF = np.random.rand(3, 3, 8, 1000)

    for a in range(3):
        FF[a, a] += 1

    # init Material
    W = MaterialTensor(x=[F], fun=lambda x: x[0])

    W0 = W.function([FF])
    dW = W.gradient([FF])
    DW = W.jacobian([FF])

    # dW and DW are always lists...
    assert W0[0].shape == (3, 3, 8, 1000)
    assert dW[0].shape == (3, 3, 3, 3, 8, 1000)
    assert DW[0].shape == (3, 3, 3, 3, 8, 1000)

    # check output of parallel evaluation
    W0 = W.function([FF], threads=2)
    dW = W.gradient([FF], threads=2)
    DW = W.jacobian([FF], threads=2)

    assert W0[0].shape == (3, 3, 8, 1000)
    assert dW[0].shape == (3, 3, 3, 3, 8, 1000)
    assert DW[0].shape == (3, 3, 3, 3, 8, 1000)

    # init Material
    pp = np.random.rand(8, 1000)
    W = MaterialTensor(x=[p], fun=lambda x: x[0], compress=True)
    W0 = W.function([pp], threads=2)

    assert W0[0].shape == (8, 1000)

    # init Material
    pp = np.random.rand(1, 1, 8, 1000)
    W = MaterialTensor(x=[p], fun=lambda x: x[0])
    W0 = W.function([pp], threads=2)

    assert W0[0].shape == (1, 1, 8, 1000)


if __name__ == "__main__":
    test_simple()
    test_tensor()



from ..._helpers import isochoric_volumetric_split
from ....math import transpose, sum1, diag, sqrt, inv, det
from ..quadrature._bazant_oh import BazantOh


@isochoric_volumetric_split
def microsphere_affine_stretch(F, f, kwargs, quadrature=BazantOh(n=21)):
    "Micro-sphere model: Affine stretch part."

    r = quadrature.points
    w = quadrature.weights

    C = transpose(F) @ F
    affine_stretch = sqrt(diag(r.T @ C @ r))

    return sum1(f(affine_stretch, **kwargs) * w)


@isochoric_volumetric_split
def microsphere_affine_tube(F, f, kwargs, quadrature=BazantOh(n=21)):
    "Micro-sphere model: Affine area-stretch part."

    r = quadrature.points
    w = quadrature.weights

    Fs = det(F) * transpose(inv(F))
    Cs = transpose(Fs) @ Fs
    affine_areastretch = sqrt(diag(r.T @ Cs @ r))

    return sum1(f(affine_areastretch, **kwargs) * w)



from matadi import MaterialHyperelastic, Lab
import matadi.models as md
from matadi.models import neo_hooke, extended_tube, van_der_waals, mooney_rivlin

import matplotlib.pyplot as plt

import pytest


def library():
    "Library with models and parameters."

    database = {
        md.saint_venant_kirchhoff: {"mu": 1.0, "lmbda": 20.0},
        md.neo_hooke: {"C10": 0.5},
        md.mooney_rivlin: {"C10": 0.3, "C01": 0.8},
        md.yeoh: {"C10": 0.5, "C20": 0.1, "C30": 0.01},
        md.third_order_deformation: {
            "C10": 0.3,
            "C01": 0.2,
            "C11": 0.02,
            "C20": -0.1,
            "C30": 0.02,
        },
        md.ogden: {"mu": (1.0, 0.2), "alpha": (2.0, -1.5)},
        md.arruda_boyce: {"C1": 1.0, "limit": 3.2},
        md.extended_tube: {"Gc": 0.1867, "Ge": 0.2169, "beta": 0.2, "delta": 0.09693},
        md.van_der_waals: {"mu": 1.0, "beta": 0.1, "a": 0.5, "limit": 5.0},
    }

    return database


def pre(
    model,
    num=20,
    test_without_bulk=False,
    stability=False,
    plot=False,
    plot_shear=False,
    close=True,
    run_all=False,
    run_kwargs={},
):

    lib = library()
    kwargs = lib[model]

    if test_without_bulk:
        # init material without bulk modulus
        mat = MaterialHyperelastic(model, **kwargs)

    # init material with bulk modulus
    mat = MaterialHyperelastic(model, **kwargs, bulk=5000.0)

    # init lab
    lab = Lab(mat)

    # run experiments
    if run_all:
        data = lab.run(ux=False, bx=False, ps=False, shear=False, num=num, **run_kwargs)
        data = lab.run(ux=True, bx=False, ps=False, shear=False, num=num, **run_kwargs)
        data = lab.run(ux=True, bx=True, ps=False, shear=False, num=num, **run_kwargs)

        data = lab.run(ux=False, bx=False, ps=False, shear=True, num=num, **run_kwargs)
        data = lab.run(ux=True, bx=False, ps=False, shear=True, num=num, **run_kwargs)
        data = lab.run(ux=True, bx=True, ps=False, shear=True, num=num, **run_kwargs)

    data = lab.run(ux=True, bx=True, ps=True, shear=True, num=num, **run_kwargs)

    if plot:
        # plot stress vs. stretch
        fig, ax = lab.plot(data, stability=stability)

        if close:
            plt.close(fig)

    if plot_shear:
        # plot stress vs. stretch
        fig, ax = lab.plot_shear(data)

        if close:
            plt.close(fig)

    # dW and DW are lists
    assert len(data[0].stress) == num
    assert len(data[0].stretch) == num

    with pytest.raises(TypeError):
        data_x = lab.run(ux=True, bx=True, ps=True, shear=False, num=num, **run_kwargs)
        lab.plot_shear(data_x)

    return data


def test_lab():

    data = pre(neo_hooke, test_without_bulk=True)
    data = pre(neo_hooke, run_all=True, plot=True, plot_shear=True, close=True)

    run_kwargs = {"stretch_min": 0.1, "stretch_max": 1.0, "shear_max": 0.8}

    data = pre(neo_hooke, run_kwargs=run_kwargs)
    data = pre(extended_tube)
    data = pre(van_der_waals)
    data = pre(mooney_rivlin, close=True, stability=True, plot=True, plot_shear=True)

    del data


if __name__ == "__main__":
    test_lab()



from ._helpers import (
    isochoric_volumetric_split,
    volumetric,
)

from ._hyperelasticity_isotropic import (
    linear_elastic,
    saint_venant_kirchhoff,
    neo_hooke,
    mooney_rivlin,
    yeoh,
    third_order_deformation,
    ogden,
    arruda_boyce,
    extended_tube,
    van_der_waals,
)

from ._hyperelasticity_anisotropic import (
    fiber,
    fiber_family,
    holzapfel_gasser_ogden,
)

from . import microsphere
from .microsphere.nonaffine import miehe_goektepe_lulei



import numpy as np

from matadi import (
    MaterialHyperelasticPlaneStrain,
    MaterialHyperelasticPlaneStressIncompressible,
)
from matadi.models import neo_hooke


def pre():

    FF = np.random.rand(2, 2, 8, 1000)
    for a in range(2):
        FF[a, a] += 1

    return FF


def test_plane_strain():

    # data
    FF = pre()

    # init Material
    W = MaterialHyperelasticPlaneStrain(
        fun=neo_hooke,
        C10=0.5,
    )

    W0 = W.function([FF])
    dW = W.gradient([FF])
    DW = W.hessian([FF])

    assert W0[0].shape == (8, 1000)
    assert dW[0].shape == (2, 2, 8, 1000)
    assert DW[0].shape == (2, 2, 2, 2, 8, 1000)

    assert W0[0].shape == (8, 1000)
    assert dW[0].shape == (2, 2, 8, 1000)
    assert DW[0].shape == (2, 2, 2, 2, 8, 1000)


def test_plane_stress():

    # data
    FF = pre()

    # init Material
    W = MaterialHyperelasticPlaneStressIncompressible(
        fun=neo_hooke,
        C10=0.5,
    )

    W0 = W.function([FF])
    dW = W.gradient([FF])
    DW = W.hessian([FF])

    assert W0[0].shape == (8, 1000)
    assert dW[0].shape == (2, 2, 8, 1000)
    assert DW[0].shape == (2, 2, 2, 2, 8, 1000)

    assert W0[0].shape == (8, 1000)
    assert dW[0].shape == (2, 2, 8, 1000)
    assert DW[0].shape == (2, 2, 2, 2, 8, 1000)


if __name__ == "__main__":
    test_plane_strain()
    test_plane_stress()



import numpy as np

from matadi import Variable, Material, MaterialTensor
from matadi.math import transpose, eigvals, sum1, trace, cof, inv, det, SX


def fun(x):
    F = x[0]
    return (sum1(eigvals(transpose(F) @ F))[0, 0] - 3) / 2


def test_eigvals():

    # variables
    F = Variable("F", 3, 3)

    # data
    np.random.seed(2345537)
    FF = np.random.rand(3, 3, 5, 100)
    for a in range(3):
        FF[a, a] += 1

    # input with repeated equal eigenvalues
    FF[:, :, 0, 1] = np.diag(2.4 * np.ones(3))
    FF[:, :, 0, 2] = np.diag([2.4, 1.2, 2.4])

    # init Material
    W = Material(x=[F], fun=fun)

    WW = W.function([FF])
    dW = W.gradient([FF])
    DW = W.hessian([FF])

    Eye = np.eye(3)
    Eye4 = np.einsum("ij,kl->ikjl", Eye, Eye)

    # check function
    f = FF[:, :, 0, 0]
    c = f.T @ f
    assert np.isclose(WW[0][0, 0], (np.linalg.eigvals(c).sum() - 3) / 2)

    # check gradient
    assert np.allclose(dW[0][:, :, 0, 0], FF[:, :, 0, 0])
    assert np.allclose(dW[0][:, :, 0, 1], FF[:, :, 0, 1])
    assert np.allclose(dW[0][:, :, 0, 2], FF[:, :, 0, 2])

    # check hessian
    assert np.allclose(DW[0][:, :, :, :, 0, 0], Eye4)
    assert np.allclose(DW[0][:, :, :, :, 0, 1], Eye4)
    assert np.allclose(DW[0][:, :, :, :, 0, 2], Eye4)


def test_eigvals_single():

    # variables
    F = Variable("F", 3, 3)

    # data
    FF = np.diag(2.4 * np.ones(3))

    # init Material
    W = Material(x=[F], fun=fun)

    WW = W.function([FF])
    WW = W.function([FF])
    dW = W.gradient([FF])
    DW = W.hessian([FF])

    Eye = np.eye(3)
    Eye4 = np.einsum("ij,kl->ikjl", Eye, Eye)

    # check function
    f = FF
    c = f.T @ f
    assert np.isclose(WW[0], (np.linalg.eigvals(c).sum() - 3) / 2)

    # check gradient
    assert np.allclose(dW[0], FF)

    # check hessian
    assert np.allclose(DW[0], Eye4)


def test_cof():

    # variables
    F = Variable("F", 3, 3)

    # data
    FF = np.diag(2.4 * np.ones(3)).reshape(3, 3, 1, 1)

    # fun
    def g(x):
        F = x[0]
        return cof(F)

    # init Material
    W = MaterialTensor(x=[F], fun=g)
    W = MaterialTensor(x=[F], fun=lambda x: x[0])

    Eye = np.eye(3)
    Eye4 = np.einsum("ij,kl->ikjl", Eye, Eye)

    WW = W.function([FF])
    dW = W.gradient([FF])
    DW = W.jacobian([FF])

    assert np.allclose(dW, DW)
    assert np.allclose(dW[0][:, :, :, :, 0, 0], Eye4)


if __name__ == "__main__":
    test_eigvals()
    test_eigvals_single()
    test_cof()



import numpy as np

from matadi import Variable, MaterialHyperelastic
from matadi.models import fiber, fiber_family, holzapfel_gasser_ogden
from matadi.math import det, transpose, trace, invariants, sqrt


def test_fiber():

    # data
    FF = np.zeros((3, 3, 2))
    for a in range(3):
        FF[a, a] += 1

    for model in [fiber, fiber_family]:

        # init Material without bulk
        M = MaterialHyperelastic(model, E=1, angle=30, axis=2)

        # init Material
        M = MaterialHyperelastic(model, E=1, angle=30, axis=2, bulk=5000)

        W0 = M.function([FF])
        dW = M.gradient([FF])
        DW = M.hessian([FF])

        assert W0[0].shape == (2,)
        assert dW[0].shape == (
            3,
            3,
            2,
        )
        assert DW[0].shape == (
            3,
            3,
            3,
            3,
            2,
        )


def test_hgo():

    # data
    FF = np.zeros((3, 3, 2))
    for a in range(3):
        FF[a, a] += 1

    for model in [holzapfel_gasser_ogden]:

        # init Material without bulk
        M = MaterialHyperelastic(
            model, c=0.0764, k1=996.6, k2=524.6, kappa=0.2, angle=49.98, axis=2
        )

        # init Material
        M = MaterialHyperelastic(
            model,
            c=0.0764,
            k1=996.6,
            k2=524.6,
            kappa=0.2,
            angle=49.98,
            axis=2,
            bulk=5000,
        )

        W0 = M.function([FF])
        dW = M.gradient([FF])
        DW = M.hessian([FF])

        assert W0[0].shape == (2,)
        assert dW[0].shape == (
            3,
            3,
            2,
        )
        assert DW[0].shape == (
            3,
            3,
            3,
            3,
            2,
        )


if __name__ == "__main__":
    test_fiber()
    test_hgo()



import numpy as np

import matadi
import matadi.models as md


def get_title(model):
    return "-".join([m.title() for m in model.__name__.split("_")])


def library():
    "Library with models and parameters."

    q = md.microsphere.quadrature.BazantOh(n=21)

    database = {
        md.saint_venant_kirchhoff: {"mu": 1.0, "lmbda": 20.0},
        md.neo_hooke: {"C10": 0.5},
        md.mooney_rivlin: {"C10": 0.3, "C01": 0.8},
        md.yeoh: {"C10": 0.5, "C20": 0.1, "C30": 0.01},
        md.third_order_deformation: {
            "C10": 0.3,
            "C01": 0.2,
            "C11": 0.02,
            "C20": -0.1,
            "C30": 0.02,
        },
        md.ogden: {"mu": (1.0, 0.2), "alpha": (2.0, -1.5)},
        md.arruda_boyce: {"C1": 1.0, "limit": 3.2},
        md.extended_tube: {"Gc": 0.1867, "Ge": 0.2169, "beta": 0.2, "delta": 0.09693},
        md.van_der_waals: {"mu": 1.0, "beta": 0.1, "a": 0.5, "limit": 5.0},
        md.microsphere.affine.stretch: {
            "quadrature": q,
            "f": md.microsphere.langevin,
            "kwargs": {"mu": 1, "N": 10},
        },
        md.microsphere.affine.tube: {
            "quadrature": q,
            "f": md.microsphere.langevin2,
            "kwargs": {"mu": 1, "N": 10},
        },
        md.microsphere.nonaffine.stretch: {
            "quadrature": q,
            "p": 2.7,
            "f": md.microsphere.gauss,
            "kwargs": {"mu": 1},
        },
        md.microsphere.nonaffine.tube: {
            "quadrature": q,
            "q": 2.7,
            "f": md.microsphere.linear,
            "kwargs": {"mu": 1},
        },
        matadi.models.miehe_goektepe_lulei: {
            "mu": 0.1475,
            "N": 3.273,
            "p": 9.31,
            "U": 9.94,
            "q": 0.567,
        },
        matadi.models.linear_elastic: {
            "mu": 1.0,
            "lmbda": 2.0,
        },
    }

    return database


def test_models():

    # data
    np.random.seed(2345537)
    FF = np.random.rand(3, 3, 5, 100)
    for a in range(3):
        FF[a, a] += 1

    pp = np.random.rand(5, 100)
    JJ = 1 + np.random.rand(5, 100) / 10

    lib = library()

    for model, kwargs in lib.items():

        if model not in [md.saint_venant_kirchhoff, md.linear_elastic]:
            kwargs["bulk"] = 5000.0

        HM = matadi.MaterialHyperelastic(model, **kwargs)
        HM_mixed = matadi.ThreeFieldVariation(HM)

        W = HM.function([FF])
        P = HM.gradient([FF], threads=1)
        A = HM.hessian([FF])

        assert len(W) == 1
        assert len(P) == 1
        assert len(A) == 1

        W = HM_mixed.function([FF, pp, JJ])
        P = HM_mixed.gradient([FF, pp, JJ])
        A = HM_mixed.hessian([FF, pp, JJ])

        assert len(W) == 1
        assert len(P) == 3
        assert len(A) == 6

    nh = matadi.MaterialHyperelastic(md.neo_hooke, **lib[md.neo_hooke])
    mr = matadi.MaterialHyperelastic(md.mooney_rivlin, **lib[md.mooney_rivlin])

    comp = matadi.MaterialComposite([nh, mr])
    comp.fun()

    W = comp.function([FF])
    P = comp.gradient([FF])
    A = comp.hessian([FF])

    assert len(W) == 1
    assert len(P) == 1
    assert len(A) == 1

    nh_mixed = matadi.ThreeFieldVariation(nh)
    mr_mixed = matadi.ThreeFieldVariation(mr)

    comp = matadi.MaterialComposite([nh_mixed, mr_mixed])

    W = comp.function([FF, pp, JJ])
    P = comp.gradient([FF, pp, JJ])
    A = comp.hessian([FF, pp, JJ])

    assert len(W) == 1
    assert len(P) == 3
    assert len(A) == 6


if __name__ == "__main__":
    test_models()
